﻿using System.Collections.Generic;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Mapping;
using com.orbitgis.tool;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    class MapObjects
    {
        // Define the CRS that will be used to communicate with the 3D Mapping SDK (WGS84 lon-lat geographic EPSG:4326)
        static internal string CRS = orbitgt.mapping3d.sdk.viewer.Constants.CRS_WGS84;

        // the EPSG code of the coordinate reference system (CRS) we use to communicate with the viewer
        static internal int VIEWER_CRS_CODE = 4326; // the well-known WGS84 system
        static internal SpatialReference VIEWER_CRS = null;
        // the EPSG code of the coordinate reference system (CRS) we use to define host maps in
        static internal int HOST_CRS_CODE = 3395; // the worldwide mercator system
        static internal SpatialReference HOST_CRS = null;

        /// <summary>
        /// Convert a point map object to a point geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal Geometry getPointGeometry(MapObject mapObject)
        {
            if (mapObject.spatialPoint == null) return null;
            return GeometryEngine.Instance.ImportFromWKT(WKTImportFlags.wktImportNonTrusted, mapObject.spatialPoint.asWKT(), VIEWER_CRS);
        }

        /// <summary>
        /// Convert a line map object to a line geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal Geometry getLineGeometry(MapObject mapObject)
        {
            if (mapObject.spatialLine == null) return null;
            return GeometryEngine.Instance.ImportFromWKT(WKTImportFlags.wktImportNonTrusted, mapObject.spatialLine.asPolylineWKT(), VIEWER_CRS);
        }

        /// <summary>
        /// Convert an area map object to an area geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal Geometry getAreaGeometry(MapObject mapObject)
        {
            if (mapObject.spatialArea == null) return null;
            return GeometryEngine.Instance.ImportFromWKT(WKTImportFlags.wktImportNonTrusted, mapObject.spatialArea.asPolygonWKT(), VIEWER_CRS);
        }

        /// <summary>
        /// Get the centroid position of a path.
        /// </summary>
        /// <param name="path">the path</param>
        /// <returns>the centroid position</returns>
        static public Coordinate GetPathCentroid(CoordinatePath path)
        {
            double tx = 0.0;
            double ty = 0.0;
            double tz = 0.0;
            int tcount = 0;
            foreach (CoordinatePart part in path.parts.toArray())
            {
                foreach (Coordinate point in part.points.toArray())
                {
                    tx += point.x;
                    ty += point.y;
                    if (point.is3D) tz += point.z;
                    tcount++;
                }
            }
            if (tcount == 0) return null;
            return new Coordinate(path.crs, tx / tcount, ty / tcount, tz / tcount, (tz != 0.0));
        }

        /// <summary>
        /// Get the centroid position of a map object.
        /// </summary>
        /// <param name="mapObject">the map object</param>
        /// <returns>the centroid position (null if not available)</returns>
        static public Coordinate GetMapObjectCentroid(MapObject mapObject)
        {
            if (mapObject.spatialPoint != null) return mapObject.spatialPoint;
            if (mapObject.spatialLine != null) return GetPathCentroid(mapObject.spatialLine);
            if (mapObject.spatialArea != null) return GetPathCentroid(mapObject.spatialArea);
            return null;
        }

        /// <summary>
        /// Get the position of a symbol path on the map view.
        /// </summary>
        /// <param name="mapObject">the symbol path object</param>
        /// <param name="mapView">the map view to display on</param>
        /// <returns>the map point (in the map spatial reference)</returns>
        static public MapPoint GetSymbolPathOrigin(MapObject mapObject, MapView mapView)
        {
            Coordinate centroid = GetMapObjectCentroid(mapObject);
            MapPoint origin = MapPointBuilder.CreateMapPoint(centroid.x, centroid.y, MapObjects.VIEWER_CRS);
            return GeometryEngine.Instance.Project(origin, mapView.Map.SpatialReference) as MapPoint;
        }

        /// <summary>
        /// Get the geometry of a symbol path on the map view
        /// </summary>
        /// <param name="mapObject">the symbol path object</param>
        /// <param name="mapView">the map view to display on</param>
        /// <returns>the map geometry (in the map spatial reference)</returns>
        static public Geometry GetSymbolPathGeometry(MapObject mapObject, MapView mapView)
        {
            /* Get the client position of the symbol origin */
            System.Windows.Point origin = mapView.MapToClient(GetSymbolPathOrigin(mapObject, mapView));
            /* Add the parts of the symbol path */
            PolygonBuilder builder = new PolygonBuilder(mapView.Map.SpatialReference);
            for (int i = 0; i < mapObject.symbolPath.parts.size(); i++)
            {
                /* Add all points of the next part */
                List<Coordinate2D> coordinates = new List<Coordinate2D>();
                CoordinatePart part = mapObject.symbolPath.parts.get(i);
                for (int j = 0; j < part.points.size(); j++)
                {
                    /* Add the next point */
                    Coordinate pixel = part.points.get(j);
                    MapPoint pointM = mapView.ClientToMap(new System.Windows.Point(origin.X + pixel.x, origin.Y + pixel.y));
                    coordinates.Add(pointM.Coordinate2D);
                }
                builder.AddPart(coordinates);
            }
            /* Return the geometry */
            return builder.ToGeometry();
        }

        /// <summary>
        /// Get the geometry of a map object.
        /// </summary>
        /// <param name="mapObject">the map object</param>
        /// <param name="mapView">the map view to display on</param>
        /// <returns>the map geometry</returns>
        static public Geometry GetMapObjectGeometry(MapObject mapObject, MapView mapView)
        {
            /* Point? */
            if (mapObject.spatialPoint != null)
            {
                /* Symbol path? */
                if (mapObject.symbolPath != null)
                {
                    return GetSymbolPathGeometry(mapObject, mapView);
                }
                /* 3D? */
                else if (mapObject.spatialPoint.is3D)
                {
                    MapPoint point = MapPointBuilder.CreateMapPoint(mapObject.spatialPoint.x, mapObject.spatialPoint.y, mapObject.spatialPoint.z, MapObjects.VIEWER_CRS);
                    return GeometryEngine.Instance.Project(point, MapObjects.HOST_CRS);
                }
                /* 2D */
                else
                {
                    MapPoint point = MapPointBuilder.CreateMapPoint(mapObject.spatialPoint.x, mapObject.spatialPoint.y, MapObjects.VIEWER_CRS); // class MapPoint
                    return GeometryEngine.Instance.Project(point, MapObjects.HOST_CRS);
                }
            }
            /* Line? */
            if (mapObject.spatialLine != null)
            {
                PolylineBuilder builder = new PolylineBuilder(MapObjects.VIEWER_CRS);
                for (int i = 0; i < mapObject.spatialLine.parts.size(); i++)
                {
                    List<Coordinate3D> coordinates = new List<Coordinate3D>();
                    CoordinatePart part = mapObject.spatialLine.parts.get(i);
                    for (int j = 0; j < part.points.size(); j++)
                    {
                        Coordinate point = part.points.get(j);
                        coordinates.Add(new Coordinate3D(point.x, point.y, point.z));
                    }
                    builder.AddPart(coordinates);
                }
                Polyline geometry = builder.ToGeometry();
                return GeometryEngine.Instance.Project(geometry, MapObjects.HOST_CRS);
            }
            /* Area? */
            if (mapObject.spatialArea != null)
            {
                //Message.WriteLine("Creating polygon ("+ mapObject.spatialArea.parts.size() + " parts) in CRS " + MapObjects.VIEWER_CRS.Wkid);
                PolygonBuilder builder = new PolygonBuilder(MapObjects.VIEWER_CRS);
                for (int i = 0; i < mapObject.spatialArea.parts.size(); i++)
                {
                    //Message.WriteLine("Adding part "+i);
                    List<Coordinate3D> coordinates = new List<Coordinate3D>();
                    CoordinatePart part = mapObject.spatialArea.parts.get(i);
                    for (int j = 0; j < part.points.size(); j++)
                    {
                        Coordinate point = part.points.get(j);
                        //Message.WriteLine("Adding point "+j+" ("+ Strings.toStringDouble(point.x) + ","+ Strings.toStringDouble(point.y) + ","+ Strings.toStringDouble(point.z) + ")");
                        coordinates.Add(new Coordinate3D(point.x, point.y, point.z));
                    }
                    builder.AddPart(coordinates);
                }
                Polygon geometry = builder.ToGeometry();
                //Message.WriteLine("Projecting to CRS "+ MapObjects.HOST_CRS.Wkid);
                return GeometryEngine.Instance.Project(geometry, MapObjects.HOST_CRS);
            }
            /* Unknown type */
            return null;
        }
    }
}
